local AddonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local c = BittensSpellFlashLibrary

local GetTime = GetTime
local IsMounted = IsMounted
local UnitExists = UnitExists
local math = math
local print = print

c.Init(a)

a.Costs = {}
local function setCost(blood, frost, unholy, name, bonusRP, chill, rime)
	a.Costs[s.SpellName(c.GetID(name))] = {
		Blood = blood,
		Frost = frost,
		Unholy = unholy,
		BonusRP = bonusRP,
		Chill = chill,
		Rime = rime,
	}
end

--setCost(-1, 0, 0, "Blood Tap")
--setCost(-2, -2, -2, "Empower Rune Weapon", 25)
setCost(0, 0, 0, "Empower Rune Weapon", 25)
setCost(0, 0, 0, "Horn of Winter", 10)
setCost(0, 0, 1, "Bone Shield")
setCost(0, 0, 1, "Dark Transformation")
setCost(0, 0, 1, "Death and Decay")
setCost(0, 0, 1, "Pestilence")
setCost(0, 0, 1, "Plague Strike")
setCost(0, 0, 1, "Scourge Strike")
setCost(0, 1, 0, "Chains of Ice", nil, true, nil)
setCost(0, 1, 0, "Howling Blast", nil, true, true)
setCost(0, 1, 0, "Icy Touch", nil, true, true)
setCost(0, 1, 0, "Pillar of Frost")
setCost(0, 1, 1, "Death Strike")
setCost(0, 1, 1, "Obliterate", nil, true, nil)
setCost(1, 0, 0, "Blood Boil")
setCost(1, 0, 0, "Heart Strike")
setCost(1, 1, 0, "Festering Strike")

------------------------------------------------------------------------ common
local function isFullyDepleted(runeID)
	local otherID
	if runeID % 2 == 0 then
		otherID = runeID - 1
	else
		otherID = runeID + 1
	end
	local otherCD = a.Runes[otherID]
	return otherCD > 0 and a.Runes[runeID] > otherCD
end

local function runeAvailable(runeId, delay)
	if delay == nil then
		delay = 0
	end
	return a.Runes[runeId] <= delay
end

function a.BothRunesAvailable(runeName, delay)
	if runeName == "Blood" then
		return runeAvailable(1, delay) and runeAvailable(2, delay)
	elseif runeName == "Frost" then
		return runeAvailable(5, delay) and runeAvailable(6, delay)
	elseif runeName == "Unholy" then
		return runeAvailable(3, delay) and runeAvailable(4, delay)
	end
	print("Bad rune name: ", runeName)
end

local function hasFrostFever(earlyRefresh)
	return c.GetMyDebuffDuration("Frost Fever") > earlyRefresh
		or c.IsAuraPendingFor("Outbreak")
		or c.IsAuraPendingFor("Icy Touch")
		or (c.HasGlyph("Howling Blast") and c.IsAuraPendingFor("Howling Blast"))
end

local function hasBloodPlague(earlyRefresh)
	return c.GetMyDebuffDuration("Blood Plague") > earlyRefresh
		or c.IsAuraPendingFor("Outbreak")
		or c.IsAuraPendingFor("Plague Strike")
end

local function hasBothDiseases()
	return hasFrostFever(0) and hasBloodPlague(0)
end

local function fakeConsume(runeID, delay, needed, deathAvailable)
	if runeAvailable(runeID, delay) then
		if a.IsDeathRune(runeID) then
			deathAvailable = deathAvailable + 1
		elseif needed > 0 then
			needed = needed - 1
		end
	end
	return needed, deathAvailable
end

local function sufficientResources(z)
	local name = s.SpellName(z.ID)
	local cost = a.Costs[name]
	if cost then
		local delay = 0
		if z.NoGCD then
			delay = -c.GetBusyTime()
		end
		local deathAvailable = 0
		local bloodNeeded = cost.Blood
		local frostNeeded = s.If(cost.Rime and a.FreezingFog > 0, 0, cost.Frost)
		local unholyNeeded = cost.Unholy
		bloodNeeded, deathAvailable = fakeConsume(
			1, delay, bloodNeeded, deathAvailable)
		bloodNeeded, deathAvailable = fakeConsume(
			2, delay, bloodNeeded, deathAvailable)
		frostNeeded, deathAvailable = fakeConsume(
			5, delay, frostNeeded, deathAvailable)
		frostNeeded, deathAvailable = fakeConsume(
			6, delay, frostNeeded, deathAvailable)
		unholyNeeded, deathAvailable = fakeConsume(
			3, delay, unholyNeeded, deathAvailable)
		unholyNeeded, deathAvailable = fakeConsume(
			4, delay, unholyNeeded, deathAvailable)
		if bloodNeeded + frostNeeded + unholyNeeded > deathAvailable then
			return false
		end
	end
--c.Debug("Resources", s.SpellCost(name), "<=", a.RP, "|", s.SpellCost(name) <= a.RP)
	return s.SpellCost(name) <= a.RP
end

local function modSpell(spell)
	spell.EvenIfNotUsable = true
	spell.NoPowerCheck = true
	spell.CheckLast = sufficientResources
end

local function addSpell(name, tag, attributes)
	modSpell(c.AddSpell(name, tag, attributes))
end

local function addOptionalSpell(name, tag, attributes)
	modSpell(c.AddOptionalSpell(name, tag, attributes))
end

c.AddSpell("Horn of Winter", nil, {
	CheckFirst = function()
		return s.PowerMissing("player") > 10
	end
})

c.AddOptionalSpell("Empower Rune Weapon", "when low", {
	NoGCD = true,
	CheckFirst = function()
		return not runeAvailable(1)
			and not runeAvailable(2)
			and not runeAvailable(3)
			and not runeAvailable(4)
			and not runeAvailable(5)
			and not runeAvailable(6)
			and s.Power("player") < 32
	end
})

c.AddSpell("Outbreak", nil, {
	CheckFirst = function()
		return not hasBothDiseases()
	end
})
c.RegisterAura("Outbreak", "Frost Fever")
c.RegisterAura("Icy Touch", "Frost Fever")
c.RegisterAura("Plague Strike", "Blood Plague")

addOptionalSpell("Death and Decay", nil, { NoRangeCheck = 1 })

c.AddOptionalSpell("Blood Tap", nil, {
	NoGCD = true,
	CheckFirst = function()
		return isFullyDepleted(1)
	end
})

modSpell(c.AddInterrupt("Mind Freeze"))

------------------------------------------------------------------------- blood
local function shouldRuneTap()
	return s.HealthPercent("player") <= 90
end

local function shouldDeathStrikeForHealth()
	return s.HealthPercent("player") <= 93
end

local function canBloodTapForDeathStrike()
	local frost = 0
	local unholy = 0
	local death = 0
	if runeAvailable(3) then
		if a.IsDeathRune(3) then
			death = death + 1
		else
			unholy = unholy + 1
		end
	end
	if runeAvailable(4) then
		if a.IsDeathRune(4) then
			death = death + 1
		else
			unholy = unholy + 1
		end
	end
	if runeAvailable(5) then
		if a.IsDeathRune(5) then
			death = death + 1
		else
			frost = frost + 1
		end
	end
	if runeAvailable(6) then
		if a.IsDeathRune(6) then
			death = death + 1
		else
			frost = frost + 1
		end
	end
	return (frost + unholy + death == 1)
		or (death == 0 and (
				(frost == 0 and unholy > 0)
			or (unholy == 0 and frost > 0)))
end

c.AddOptionalSpell("Blood Presence", nil, { Type = "form" })

c.AddSpell("Outbreak", "for Debuffs", {
	CheckFirst = function()
		local window = c.GetBusyTime() + 3
		return s.DebuffDuration(c.SLOW_MELEE_DEBUFFS) < window
			or s.DebuffDuration(c.PHYSICAL_DAMAGE_DEBUFFS) < window
	end
})

addSpell("Icy Touch", "for Debuff", {
	Debuff = c.SLOW_MELEE_DEBUFFS,
	CheckFirst = function()
		return c.GetCooldown("Outbreak") > 3
	end
})

addSpell("Plague Strike", "for Debuff", {
	Debuff = c.PHYSICAL_DAMAGE_DEBUFFS,
	CheckFirst = function()
		return c.GetCooldown("Outbreak") > 3
	end
})

addSpell("Heart Strike", nil, {
	CheckFirst = function()
		if runeAvailable(1) then
			return not a.IsDeathRune(1) and runeAvailable(2, 1.5)
		end
		return runeAvailable(2)
			and not a.IsDeathRune(2)
			and runeAvailable(1, 1.5)
	end
})

addSpell("Death Strike", "for Health", {
	CheckFirst = function()
		return shouldDeathStrikeForHealth()
	end
})

addSpell("Death Strike", "for Runic Empowerment", {
	CheckFirst = function()
		return not ( -- keep at least one rune fully depleted
			isFullyDepleted(3) 
			or isFullyDepleted(4) 
			or isFullyDepleted(5) 
			or isFullyDepleted(6))
	end
})

addSpell("Death Strike", "for Rune Cooldown", {
	CheckFirst = function()
		return a.BothRunesAvailable("Frost") or a.BothRunesAvailable("Unholy")
	end
})

addSpell("Death Strike", "to Save Shield", {
	CheckFirst = function()
		return c.GetBuffDuration("Blood Shield") < 2
	end
})

addSpell("Death Strike", "if Two Available", {
	CheckFirst = function()
		local avail = 0
		if runeAvailable(3, 1.5) then
			avail = avail + 1
		end
		if runeAvailable(4, 1.5) then
			avail = avail + 1
		end
		if runeAvailable(5, 1.5) then
			avail = avail + 1
		end
		if runeAvailable(6, 1.5) then
			avail = avail + 1
		end
		return avail >= 4
	end
})

addSpell("Rune Strike", "for Runic Empowerment", {
	CheckFirst = function()
		return isFullyDepleted(3)
			or isFullyDepleted(4)
			or isFullyDepleted(5)
			or isFullyDepleted(6)
	end
})

addSpell("Rune Strike", "if Capped", {
	CheckFirst = function()
		return s.PowerMissing("player") < 20
	end
})

c.AddOptionalSpell("Rune Tap", nil, {
	FlashColor = "yellow",
	CheckFirst = function()
		return shouldRuneTap()
	end
})

addOptionalSpell("Bone Shield", nil, {
	CheckFirst = function()
		if s.InCombat() then
			return not c.HasBuff("Bone Shield")
		else
			return c.GetBuffDuration("Bone Shield") < 2 * 60
				or s.BuffStack(c.GetID("Bone Shield"), "player") < 6
		end
	end
})

addOptionalSpell("Dancing Rune Weapon")

c.AddOptionalSpell("Vampiric Blood", nil, {
	CheckFirst = function()
		return s.HealthPercent("player") < 90
	end
})

c.AddOptionalSpell("Icebound Fortitude")

c.AddOptionalSpell("Blood Tap", "for Death Strike for Health", {
	NoGCD = true,
	CheckFirst = function()
		return shouldDeathStrikeForHealth() and canBloodTapForDeathStrike()
	end
})

addSpell("Blood Boil", "under Crimson Scourge", {
	CheckFirst = function()
		return s.PowerMissing("player") > 10
			and c.HasBuff("Crimson Scourge")
	end
})

c.AddTaunt("Dark Command", nil, { NoGCD = true })

c.AddTaunt("Death Grip", nil, { NoGCD = true })

c.AddOptionalSpell("Fire of the Deep", nil, {
	Type = "item",
	Buff = "Elusive",
})

------------------------------------------------------------------------- frost
c.AddOptionalSpell("Unholy Presence", nil, { Type = "form" })

addSpell("Frost Strike")

addSpell("Frost Strike", "under KM", {
	CheckFirst = function()
		return a.KillingMachine
	end
})

addSpell("Frost Strike", "under KM no Diseases", {
	CheckFirst = function()
		return a.KillingMachine and not hasBothDiseases()
	end
})

c.RegisterAura("Howling Blast", "Frost Fever")

addSpell("Howling Blast")

addSpell("Howling Blast", "for FF", {
	CheckFirst = function()
		return c.HasGlyph("Howling Blast") and not hasFrostFever(0)
	end
})

addSpell("Plague Strike", "for BP unless Outbreak", {
	CheckFirst = function()
		return c.GetCooldown("Outbreak") > 3 and not hasBloodPlague(0)
	end
})

addSpell("Obliterate", nil, {
	CheckFirst = function()
		return hasBothDiseases()
	end
})

addSpell("Obliterate", "UU", {
	CheckFirst = function()
		return a.BothRunesAvailable("Unholy", 1) and hasBothDiseases()
	end
})

addSpell("Obliterate", "under KM no Freezing Fog", {
	CheckFirst = function()
		return a.KillingMachine
			and a.FreezingFog == 0
			and hasBothDiseases()
	end
})

addOptionalSpell("Pillar of Frost", nil, { NoGCD = true })

addSpell("Howling Blast", "BB or FF", {
	CheckFirst = function()
		return a.BothRunesAvailable("Blood", 1)
			or a.BothRunesAvailable("Frost", 1)
	end
})

c.CloneSpell("Death and Decay", "UU", {
	CheckFirst = function()
		return a.BothRunesAvailable("Unholy", 1)
	end
})

addSpell("Plague Strike", "UU", {
	CheckFirst = function()
		return a.BothRunesAvailable("Unholy", 1)
	end
})

------------------------------------------------------------------------ Unholy
c.AddOptionalSpell("Raise Dead", nil, {
	NoRangeCheck = 1,
	CheckFirst = function()
		return not UnitExists("pet") and not IsMounted()
	end
})

c.AddSpell("Unholy Frenzy", nil, {
	NoRangeCheck = 1,
	Melee = 1,
	CheckFirst = function()
		return not s.Buff(c.BLOODLUST_BUFFS, "player")
	end
})

addOptionalSpell("Summon Gargoyle", nil, {
	CheckFirst = function()
		return s.BuffDuration(c.GetID("Dark Transformation"), "pet")
			- s.GlobalCooldown() > 10
	end
})

addSpell("Outbreak", "Refresh", {
	CheckFirst = function()
		return not hasFrostFever(3) or not hasBloodPlague(3)
	end
})

addSpell("Icy Touch", "Refresh", {
	CheckFirst = function()
		return not hasFrostFever(3)
	end
})

addSpell("Plague Strike", "Refresh", {
	CheckFirst = function()
		return not hasBloodPlague(3)
	end
})

addSpell("Dark Transformation", nil, {
	CheckFirst = function()
		if s.Buff(c.GetID("Dark Transformation"), "pet")
			or c.IsAuraPendingFor("Dark Transformation") then
			
			return false
		end
		
		local stacks = s.BuffStack(c.GetID("Shadow Infusion"), "pet")
		if a.ShadowInfusionPending or c.IsCasting("Death Coil") then
			stacks = stacks + 1
		end
		return stacks >= 5
	end
})

addSpell("Death Coil")

c.RegisterAura("Death Coil", "Shadow Infusion")

addSpell("Death Coil", "for Shadow Infusion", {
	CheckFirst = function()
		return not s.Buff(
				c.GetID("Dark Transformation"), "pet", c.GetBusyTime())
	end
})

addSpell("Death Coil", "under Sudden Doom if DCisOK", {
	CheckFirst = function()
		return a.DCisOK
			and c.HasBuff("Sudden Doom")
			and not c.IsCasting("Death Coil")
	end
})

local function DB_or_DF_or_UU()
	return a.BothRunesAvailable("Unholy", 1)
		or (a.IsDeathRune(1) and runeAvailable(1, 0) and runeAvailable(2, 1))
		or (a.IsDeathRune(2) and runeAvailable(2, 0) and runeAvailable(1, 1))
		or (a.IsDeathRune(5) and runeAvailable(5, 0) and runeAvailable(6, 1))
		or (a.IsDeathRune(6) and runeAvailable(6, 0) and runeAvailable(5, 1))
end

c.CloneSpell("Death and Decay", "DB or DF or UU", {
	CheckFirst = DB_or_DF_or_UU,
})

addSpell("Scourge Strike")

addSpell("Scourge Strike", "DB or DF or UU", {
	CheckFirst = DB_or_DF_or_UU,
})

addSpell("Festering Strike")

addSpell("Festering Strike", "BBFF", {
	CheckFirst = function()
		return a.BothRunesAvailable("Blood", 1)
			or a.BothRunesAvailable("Frost", 1)
	end
})
